-- **********************************************************************
-- *								        *
-- * This file was created by Krzysztof Pozniak (pozniak@ise.pw.edu.pl) *
-- * Copyright (c) 2000-2015 by Krzysztof Pozniak		        *
-- * All Rights Reserved 					        *
-- *								        *
-- **********************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

entity rpc_sender is
  generic (
    LPM_PART_WIDTH		:positive := 8;
    LPM_PART_NUM		:positive := 4;
    LPM_CHECK_WIDTH		:positive := 8;
    INPUT_REGISTERED		:boolean := true;
    OUTPUT_REGISTERED		:boolean := true
  );
  port(
    resetN			:in  std_logic;
    clock			:in  std_logic;
    part_ena			:in  std_logic;
    check_ena			:in  std_logic;
    check_data			:in  std_logic_vector(LPM_CHECK_WIDTH-1 downto 0);
    check_data_ena		:in  std_logic;
    in_data			:in  std_logic_vector(LPM_PART_WIDTH*LPM_PART_NUM-LPM_CHECK_WIDTH-1 downto 0);
    out_data			:out std_logic_vector(LPM_PART_WIDTH*LPM_PART_NUM-1 downto 0);
    test_data			:in  std_logic_vector(LPM_PART_WIDTH*LPM_PART_NUM-1 downto 0);
    test_ena			:in  std_logic;
    test_rand_ena		:in  std_logic
  );
end rpc_sender;

architecture behaviour of rpc_sender is

  function min(a, b :integer) return integer is begin
    if (a<b) then return(a); else return(b); end if;
  end function;
  --
  function max(a, b :integer) return integer is begin
    if (a>b) then return(a); else return(b); end if;
  end function;

  function part_num(vlen, plen :natural) return positive is begin
    return(((vlen-1)/plen)+1);
  end function;
  --
  function  div(v,r :natural) return natural is
    variable res :natural := v/r;
  begin
    if ((v-res*r)<((res+1)*r-v)) then return(res); else return (res+1); end if;
  end function;
  --
  function cut(src :std_logic_vector; len :positive) return std_logic_vector is
    variable res :std_logic_vector(src'length-1 downto 0) := src;
  begin
    return(res(len-1 downto 0));
  end function;
  --
  function create(len :positive; set: std_logic) return std_logic_vector is
    variable res :std_logic_vector(len-1 downto 0) := (others => set);
  begin
    return(res);
  end function;
  --
  function create(len :positive; set: std_logic_vector) return std_logic_vector is
    variable res :std_logic_vector(len-1 downto 0) := (others => '0');
  begin
    res(set'length-1 downto 0) := set;
    return(res);
  end function;
  --
  function resize(src :std_logic_vector; len :positive; set: std_logic) return std_logic_vector is
    variable res :std_logic_vector(max(src'length,len)-1 downto 0) := (others => set);
  begin
    res(src'length-1 downto 0) := src;
    return(res(len-1 downto 0));
  end function;
  --
  function rot(src :std_logic_vector; step :integer) return std_logic_vector is
    variable var :std_logic_vector(src'length-1 downto 0) := src;
    variable res :std_logic_vector(src'length-1 downto 0);
  begin
    for i in res'range loop res((i+step) mod res'length) := var(i); end loop;
    return(res);
  end function;
  --
  function part_get(src :std_logic_vector; plen :positive; index: natural) return std_logic_vector is
    variable var :std_logic_vector(src'length-1 downto 0) := src;
    variable res :std_logic_vector(plen-1 downto 0);
  begin
    res := var((index+1)*plen-1 downto index*plen);
    return(res);
  end function;
  --
  function part_set(src :std_logic_vector; index: natural; set: std_logic_vector) return std_logic_vector is
    variable res :std_logic_vector(src'length-1 downto 0) := src;
    variable var :std_logic_vector(src'length-1 downto 0) := set;
  begin
    res((index+1)*src'length-1 downto index*src'length) := set;
    return(res);
  end function;
  --
  constant DATA_WIDTH		:natural := LPM_PART_WIDTH*LPM_PART_NUM;
  constant DATA_CHECK_WIDTH	:natural := DATA_WIDTH-LPM_CHECK_WIDTH;
  constant MUX_CHECK_WIDTH	:natural := part_num(DATA_CHECK_WIDTH,LPM_CHECK_WIDTH);
  --
  constant CHECK_PART_NUM	:natural := max(div(LPM_PART_NUM,LPM_CHECK_WIDTH),1);
  constant CHECK_PART_WIDTH	:natural := LPM_PART_WIDTH*CHECK_PART_NUM;
  constant CHECK_PART_SIZE	:natural := max(div(LPM_CHECK_WIDTH,LPM_PART_NUM),1);
  constant CHECK_PART_REP	:natural := min(LPM_CHECK_WIDTH/CHECK_PART_SIZE,min(part_num(LPM_PART_NUM,CHECK_PART_NUM),LPM_CHECK_WIDTH)-1);
  constant CHECK_LPART_NUM	:natural := LPM_PART_NUM-CHECK_PART_REP*CHECK_PART_NUM;
  constant CHECK_LPART_WIDTH	:natural := LPM_PART_WIDTH*CHECK_LPART_NUM;
  constant CHECK_LPART_SIZE	:natural := LPM_CHECK_WIDTH-CHECK_PART_REP*CHECK_PART_SIZE;
  --
  signal RndReg			:std_logic_vector(DATA_WIDTH-1 downto 0);
  signal DataInReg		:std_logic_vector(DATA_CHECK_WIDTH-1 downto 0);
  signal CheckInReg		:std_logic_vector(max(LPM_CHECK_WIDTH,1)-1 downto 0);
  signal DataSig		:std_logic_vector(DATA_WIDTH-1 downto 0);
  signal DataReg		:std_logic_vector(DATA_WIDTH-1 downto 0);

begin

  process (clock, resetN, check_data, in_data)
  begin
    if (INPUT_REGISTERED=FALSE) then
      DataInReg  <= in_data;
      CheckInReg <= check_data;
    elsif (resetN='0') then
      DataInReg  <= (others=>'0');
      CheckInReg <= (others=>'0');
    elsif (clock'event and clock='1') then
      DataInReg  <= in_data;
      CheckInReg <= check_data;
    end if;
  end process;
  --
  process (DataInReg, CheckInReg, RndReg, test_data, part_ena, check_ena, check_data_ena, test_ena, test_rand_ena)
    variable DataMoveVar  :std_logic_vector(DATA_WIDTH-1 downto 0);
    variable DataCheckVar :std_logic_vector(DATA_WIDTH-1 downto 0);
  begin
    if (test_ena='0') then
      if (LPM_CHECK_WIDTH=0) then
        DataMoveVar := DataInReg;
      elsif (check_data_ena='1') then
        DataMoveVar := cut(CheckInReg & DataInReg,DATA_WIDTH);
      else
        DataMoveVar := cut(create(CheckInReg'length,'0') & DataInReg,DATA_WIDTH);
      end if;
    else
      if (test_rand_ena='0') then
        DataMoveVar := test_data;
        if (LPM_CHECK_WIDTH>0 and check_data_ena='1') then
          DataMoveVar := DataMoveVar xor create(DataMoveVar'length,CheckInReg);
          if (CheckInReg'length>1) then
            DataMoveVar := DataMoveVar xor rot(create(DataMoveVar'length,CheckInReg),CheckInReg'length/2);
          end if;
        end if;
      else
        DataMoveVar := RndReg;
        if (LPM_CHECK_WIDTH>0 and check_data_ena='1') then
          for index in 0 to LPM_PART_NUM-1 loop
            DataMoveVar(LPM_PART_WIDTH*(index+1)-1 downto LPM_PART_WIDTH*index)
            := DataMoveVar(LPM_PART_WIDTH*(index+1)-1 downto LPM_PART_WIDTH*index) xor resize(CheckInReg,LPM_PART_WIDTH,'0');
          end loop;
        end if;
      end if;
    end if;
    DataCheckVar := DataMoveVar;
    if (LPM_CHECK_WIDTH > 0 and check_ena='1') then
      for index in 0 to CHECK_PART_REP-1 loop
        DataCheckVar(DATA_CHECK_WIDTH+CHECK_PART_SIZE*(index+1)-1 downto DATA_CHECK_WIDTH+CHECK_PART_SIZE*index)
        :=  DataCheckVar(DATA_CHECK_WIDTH+CHECK_PART_SIZE*(index+1)-1 downto DATA_CHECK_WIDTH+CHECK_PART_SIZE*index)
        xor create(CHECK_PART_SIZE,XOR_REDUCE(part_get(DataCheckVar(DATA_CHECK_WIDTH-1 downto 0),CHECK_PART_WIDTH-CHECK_PART_SIZE,index)));
      end loop;
      DataCheckVar(DATA_WIDTH-1 downto DATA_WIDTH-CHECK_LPART_SIZE)
      :=  DataCheckVar(DATA_WIDTH-1 downto DATA_WIDTH-CHECK_LPART_SIZE)
      xor create(CHECK_LPART_SIZE,XOR_REDUCE(DataCheckVar(DATA_CHECK_WIDTH-1 downto DATA_CHECK_WIDTH-CHECK_LPART_WIDTH+CHECK_LPART_SIZE)));
    end if;
    DataMoveVar := DataCheckVar;
    if (part_ena='1' and test_ena='0') then
      if (LPM_CHECK_WIDTH = 0) then
        for y in 0 to LPM_PART_WIDTH-1 loop
          for x in 0 to LPM_PART_NUM-1 loop
            DataMoveVar((LPM_PART_WIDTH*x+y)) := DataCheckVar((LPM_PART_NUM*y+x));
          end loop;
        end loop;
      else
        for rep in 0 to CHECK_PART_REP-1 loop
	  DataMoveVar := part_set(DataMoveVar,rep,   (part_get(DataCheckVar(DATA_WIDTH-1 downto DATA_WIDTH-LPM_CHECK_WIDTH),CHECK_PART_SIZE,rep)
	                                             & part_get(DataCheckVar(DATA_CHECK_WIDTH-1 downto 0),CHECK_PART_WIDTH-CHECK_PART_SIZE,rep)));
	end loop;
	DataMoveVar(DATA_WIDTH-1 downto DATA_WIDTH-CHECK_LPART_WIDTH)
	:=   DataCheckVar(DATA_WIDTH-1 downto DATA_WIDTH-CHECK_LPART_SIZE)
           & DataCheckVar(DATA_CHECK_WIDTH-1 downto DATA_CHECK_WIDTH-CHECK_LPART_WIDTH+CHECK_LPART_SIZE);
      end if;
    end if;
    DataSig <= DataMoveVar;
  end process;
  --
  process (clock, resetN, DataSig)
  begin
    if (OUTPUT_REGISTERED=TRUE and resetN='0') then
      DataReg  <= (others=>'0');
    elsif (OUTPUT_REGISTERED=FALSE or (clock'event and clock='1')) then
      DataReg  <= DataSig;
    end if;
  end process;
  out_data <= DataReg;
  --
  process (resetN, clock) is
    function get_rnd(rnd :std_logic_vector) return std_logic_vector is
      constant LEN :natural := rnd'length;
      variable RndVar  :std_logic_vector(LEN-1 downto 0);
    begin
      RndVar := rnd;
      RndVar(LEN/2) := RndVar(0) xor not(RndVar(LEN-1));
      RndVar(LEN-1 downto 0) := (RndVar(LEN-2 downto 0) & RndVar(LEN-1)) xor rnd;
      return (RndVar); 
    end function;
  begin
    if (resetN='0') then
      RndReg <= (others => '0');
    elsif (clock'event and clock='1') then
      if (test_rand_ena='0') then
        RndReg <= test_data;
      else
        for index in 0 to LPM_PART_NUM-1 loop
          RndReg(LPM_PART_WIDTH*(index+1)-1 downto LPM_PART_WIDTH*index) <= get_rnd(RndReg(LPM_PART_WIDTH*(index+1)-1 downto LPM_PART_WIDTH*index));
        end loop;
      end if;
    end if;
  end process;

end behaviour;
